IMPLEMENTATION MODULE ChipIO;      (* Allgemeine Version *)

  FROM SYSTEM IMPORT ADR, SIZE, TSIZE;
  FROM Storage IMPORT ALLOCATE, DEALLOCATE;
  IMPORT Terminal, InOut;

  CONST Zeilen       = 25;
        Spalten      = 80;
        BufLen       = 80;

        (* IBM-Grafikzeichen fuer Doppel-Rahmen *)
        LinksOben    = 311C;
        RechtsOben   = 273C;
        LinksUnten   = 310C;
        RechtsUnten  = 274C;
        WaagrechtO   = 315C;
        WaagrechtU   = 315C;
        SenkrechtL   = 272C;
        SenkrechtR   = 272C;

  TYPE Screen    = ARRAY [1..Zeilen], [1..Spalten] OF CARDINAL;
       ScreenPT  = POINTER TO Screen;
       MemArray  = ARRAY [0..Zeilen*Spalten-1] OF CARDINAL;
       MemPT     = POINTER TO MemArray;
       WindPtr   = POINTER TO WindInh;
       WindInh   = RECORD
                     x1,y1,x2,y2 : CARDINAL;  (* Fenster-Koordinaten      *)
                     cx, cy      : CARDINAL;  (* letzte Cursor-Position   *)
                     LastAttr    : CARDINAL;
                     ScrPtr      : MemPT      (* Zwischenspeicher         *)
                   END;

(* Fr Logitech etc. *)
  VAR ScreenMode [40H:49H] : BITSET;
      MonoScreen [0B000H:0]: Screen;
      ColorScreen[0B800H:0]: Screen;

(* Fr Taylor etc.
  VAR ScreenModePtr : POINTER TO BITSET; *)

  VAR ActScreen            : ScreenPT;
      LastChar             : CHAR;      (* Gepuffertes Zeichen *)
      Color                : CARDINAL;
      CurX, CurY           : CARDINAL;  (* absolute Cursor-Koordinaten *)
      MinX, MinY,
      MaxX, MaxY           : CARDINAL;  (* Fenster-Koordinaten *)
      BufPos, BufEnd       : CARDINAL;  (* Puffer fuer Line-Input *)
      Buffer               : ARRAY [0..BufLen] OF CHAR;
      WindowArray          : ARRAY [0..255] OF WindPtr;
      Screencopy           : Screen;
      TempMem              : MemArray;

  (* Low-Level Bildschirm-Routinen *)

  PROCEDURE SetCursor(X, Y : CARDINAL);  (* Setzt ANSI.SYS voraus! *)
    BEGIN
      Terminal.Write(33C); (* ESC *)
      Terminal.Write('[');
      Terminal.Write(CHR(ORD('0')+Y DIV 10));
      Terminal.Write(CHR(ORD('0')+Y MOD 10));
      Terminal.Write(';');
      Terminal.Write(CHR(ORD('0')+X DIV 10));
      Terminal.Write(CHR(ORD('0')+X MOD 10));
      Terminal.Write('H');
    END SetCursor;

  PROCEDURE Beep;
    BEGIN
      Terminal.Write(07C);
    END Beep;

  (* Standard-Modula, z.B. Taylor

  PROCEDURE KeyPressed() : BOOLEAN;
    BEGIN
      IF LastChar<>0C
      THEN
        RETURN TRUE
      ELSE
        Terminal.BusyRead(LastChar);
        RETURN LastChar<>0C
      END
    END KeyPressed;

  PROCEDURE ReadKey() : CHAR;
    VAR C : CHAR;
    BEGIN
      IF LastChar<>0C
      THEN
        C:=LastChar;
        LastChar:=0C
      ELSE
        REPEAT
          Terminal.BusyRead(C)
        UNTIL C<>0C
      END;
      RETURN C
    END ReadKey;  *)

  (* Bei Logitech: *)

  PROCEDURE KeyPressed() : BOOLEAN;
    BEGIN
      RETURN Terminal.KeyPressed()
    END KeyPressed;

  PROCEDURE ReadKey() : CHAR;
    VAR C : CHAR;
    BEGIN
      Terminal.Read(C);
      IF C=InOut.EOL THEN C:=15C END;
      IF C=0C THEN
        Terminal.Read(C);
        CASE C OF
          110C : C:=05C  (* Cursor hoch *)
        | 120C : C:=30C  (* Cursor tief *)
        | 113C : C:=23C  (* Cursor links *)
        | 115C : C:=04C  (* Cursor rechts *)
        | 111C : C:=22C  (* PageUp *)
        | 121C : C:=03C  (* PageDown *)
          ELSE C:=0C
        END (* CASE *)
      END; (* IF *)
      RETURN C
    END ReadKey;

  (* ab hier wieder fuer alle *)

  PROCEDURE Move(s, d : MemPT; l : CARDINAL);
    VAR i : CARDINAL;
    BEGIN
      FOR i:=0 TO l-1 DO d^[i]:=s^[i] END
    END Move;

  (*

    Leider arbeiten die obigen Prozeduren sehr langsam!
    Das ist kaum zu vermeiden, wenn man sich auf Standard-
    Modula beschraenkt. Falls moeglich sollten deshalb die
    folgenden effizienteren Prozeduren genommen werden:


  PROCEDURE SetCursor(X, Y : CARDINAL);
    VAR R : SYSTEM.Registers;
    BEGIN
      R.AH:=2; R.BH:=0;
      R.DH:=SHORTCARD(Y-1); R.DL:=SHORTCARD(X-1);
      Lib.Intr(R,10H);
    END SetCursor;

  PROCEDURE Beep;
    VAR R : SYSTEM.Registers;
    BEGIN
      R.AH:=14; R.AL:=7;
      Lib.Intr(R,10H);
    END Beep;

  PROCEDURE KeyPressed() : BOOLEAN;
    VAR R : SYSTEM.Registers;
    BEGIN
      IF LastChar<>0C
      THEN RETURN TRUE
      ELSE
        R.AH:=1;
        Lib.Intr(R,16H);
        RETURN SYSTEM.ZeroFlag IN R.Flags
      END
    END KeyPressed;

  PROCEDURE ReadKey() : CHAR;
    VAR C : CHAR;
        R : SYSTEM.Registers;
    BEGIN
      IF LastChar<>0C
      THEN
        C:=LastChar;
        LastChar:=0C
      ELSE
        R.AH:=0;
        Lib.Intr(R,16H);
        C:=CHR(R.AL);
        IF C=0C THEN LastChar:=CHR(R.AH); C:=33C END
      END;
      RETURN C
    END ReadKey;

  PROCEDURE Move(s, d : MemPT; l : CARDINAL);
    BEGIN
      Lib.WordMove(s, d, l)
    END Move;

  *)

  (* Scrolling-Prozeduren *)

  PROCEDURE MoveUp(X1, Y1, X2, Y2 : CARDINAL);
    VAR Y : CARDINAL;
    BEGIN
      FOR Y:=Y1 TO Y2-1 DO
        Move(ADR(ActScreen^[Y+1,X1]),ADR(ActScreen^[Y,X1]),1+X2-X1)
      END;
      FOR Y:=X1 TO X2 DO ActScreen^[Y2,Y]:=Color END
    END MoveUp;

  PROCEDURE MoveDown(X1, Y1, X2, Y2 : CARDINAL);
    VAR Y : CARDINAL;
    BEGIN
      FOR Y:=Y2 TO Y1+1 BY -1 DO
        Move(ADR(ActScreen^[Y-1,X1]),ADR(ActScreen^[Y,X1]),1+X2-X1)
      END;
      FOR Y:=X1 TO X2 DO ActScreen^[Y1,Y]:=Color END
    END MoveDown;

  (* Die universale Low-Level-Write-Prozedur,
     die die Window-Koordinaten (MinX, MinY, MaxX, MaxY)
     beachtet. Der Cursor muss nachher unbedingt mit
     SetCursor(CurX, CurY) gesetzt werden *)

  PROCEDURE WriteChar(C : CHAR);
    BEGIN
      CASE C OF
        07C : Beep                                  (* Bell            *)
      | 10C : IF CurX>MinX THEN DEC(CurX) END       (* Back Space      *)
      | 12C : IF CurY<MaxY                          (* Line Feed       *)
              THEN INC(CurY)
              ELSE MoveUp(MinX,MinY,MaxX,MaxY) END
      | 15C : CurX:=MinX                            (* Carriage Return *)
      ELSE
        ActScreen^[CurY,CurX]:=Color+ORD(C);        (* Printable Char  *)
        IF CurX<MaxX
        THEN
          INC(CurX)
        ELSE
          CurX:=MinX;
          IF CurY<MaxY
          THEN INC(CurY)
          ELSE MoveUp(MinX,MinY,MaxX,MaxY) END
        END
      END
    END WriteChar;

  (* Formatierende Write-Prozeduren fuer:
     CHAR, STRING, CARDINAL, INTEGER und REAL *)

  PROCEDURE WriteLn;
    BEGIN
      Done:=TRUE;
      WriteChar(15C);
      WriteChar(12C);
      SetCursor(CurX, CurY)
    END WriteLn;

  PROCEDURE WriteString(s : ARRAY OF CHAR);
    VAR n : CARDINAL;
    BEGIN
      Done:=TRUE;
      n:=0;
      WHILE (n<=HIGH(s)) AND (s[n]<>0C) DO
        WriteChar(s[n]);
        INC(n)
      END;
      SetCursor(CurX, CurY)
    END WriteString;

  PROCEDURE WriteCard(x, n : CARDINAL);
    VAR s : ARRAY [0..5] OF CHAR;
        i, j : CARDINAL;
    BEGIN
      Done:=TRUE;
      i:=0; j:=x;
      REPEAT
        s[i]:=CHR(ORD('0')+j MOD 10);
        j:=j DIV 10;
        INC(i)
      UNTIL (j=0) OR (i>5);
      FOR j:=i+1 TO n DO WriteChar(' ') END;
      FOR j:=i-1 TO 0 BY -1 DO WriteChar(s[j]) END;
      SetCursor(CurX, CurY)
    END WriteCard;

  PROCEDURE WriteInt(x : INTEGER; n : CARDINAL);
    VAR s : ARRAY [0..6] OF CHAR;
        i, j : CARDINAL;
    BEGIN
      Done:=TRUE;
      i:=0;
      IF x<>MIN(INTEGER)
      THEN
        j:=CARDINAL(ABS(x))
      ELSE
        j:=1+CARDINAL(MAX(INTEGER))
      END;
      REPEAT
        s[i]:=CHR(ORD('0')+j MOD 10);
        j:=j DIV 10;
        INC(i)
      UNTIL (j=0) OR (i>5);
      IF x<0 THEN s[i]:='-'; INC(i) END;
      FOR j:=i+1 TO n DO WriteChar(' ') END;
      FOR j:=i-1 TO 0 BY -1 DO WriteChar(s[j]) END;
      SetCursor(CurX, CurY)
    END WriteInt;

  PROCEDURE WriteReal(x : REAL; m, n : CARDINAL);
    VAR t : REAL;
        i, j, k : CARDINAL;
    BEGIN
      Done:=TRUE;
      t:=ABS(x);
      FOR i:=1 TO n DO t:=t*10.0 END;
      t:=t+0.5;
      j:=1;
      WHILE (t>=10.0) OR (j<=n) DO t:=t/10.0; INC(j) END;
      k:=j;
      IF n<>0 THEN INC(k) END;
      IF x<0.0 THEN INC(k) END;
      FOR i:=k+1 TO m DO WriteChar(' ') END;
      IF x<0.0 THEN WriteChar('-') END;
      FOR i:=n+1 TO j DO
        WriteChar(CHR(ORD('0')+TRUNC(t)));
        t:=(t-FLOAT(TRUNC(t)))*10.0
      END;
      IF n<>0 THEN WriteChar('.') END;
      FOR i:=1 TO n DO
        WriteChar(CHR(ORD('0')+TRUNC(t)));
        t:=(t-FLOAT(TRUNC(t)))*10.0
      END;
      SetCursor(CurX, CurY)
    END WriteReal;

  PROCEDURE Write(c : CHAR);
    BEGIN
      Done:=TRUE;
      WriteChar(c);
      SetCursor(CurX, CurY)
    END Write;

  (* Line-Input-Prozedur:
     Es wird jeweils eine Zeile eingelesen, wobei
     diese Zeile noch editiert werden kann. Diese
     Zeile wird dann Zeichen fuer Zeichen ausgegeben *)

  PROCEDURE ReadChar() : CHAR;
    VAR c : CHAR;
    BEGIN
      IF BufPos<BufEnd
      THEN
        c:=Buffer[BufPos];
        INC(BufPos);
        RETURN c
      ELSE
        BufEnd:=0;
        REPEAT
          c:=ReadKey();
          IF (c=10C) AND (BufEnd<>0)
          THEN
            DEC(BufEnd);
            WriteChar(10C);
            WriteChar(' ');
            Write(10C)
          ELSIF (c>=' ') AND (BufEnd<BufLen)
          THEN
            Buffer[BufEnd]:=c;
            INC(BufEnd);
            Write(c)
          END
        UNTIL c=15C;
        Buffer[BufEnd]:=15C;
        INC(BufEnd);
        WriteLn;
        BufPos:=1;
        RETURN Buffer[0]
      END
    END ReadChar;

  (* Read-Prozeduren fuer:
     CHAR, STRING, CARDINAL, INTEGER, REAL *)

  PROCEDURE ReadLn;
    BEGIN
      Done:=TRUE;
      BufPos:=BufEnd
    END ReadLn;

  PROCEDURE ReadString(VAR s : ARRAY OF CHAR);
    VAR c : CHAR;
        i : CARDINAL;
    BEGIN
      Done:=TRUE;
      REPEAT
        c:=ReadChar()
      UNTIL c>' ';
      i:=0;
      REPEAT
        s[i]:=c;
        INC(i);
        IF i>HIGH(s) THEN RETURN END;
        c:=ReadChar()
      UNTIL c<=' ';
      s[i]:=0C
    END ReadString;

  PROCEDURE ReadCard(VAR x : CARDINAL);
    VAR s : ARRAY [0..5] OF CHAR;
        i : CARDINAL;
    BEGIN
      ReadString(s);
      Done:=FALSE;
      x:=0;
      i:=0;
      REPEAT
        IF (i>4) OR (s[i]<'0') OR (s[i]>'9') THEN RETURN END;
        IF (x>MAX(CARDINAL) DIV 10) OR
           (10*x>MAX(CARDINAL)-(ORD(s[i])-ORD('0'))) THEN RETURN END;
        x:=10*x+(ORD(s[i])-ORD('0'));
        INC(i)
      UNTIL s[i]=0C;
      Done:=TRUE
    END ReadCard;

  PROCEDURE ReadInt(VAR x : INTEGER);
    VAR s : ARRAY [0..6] OF CHAR;
        i, k : CARDINAL;
        n : BOOLEAN;
    BEGIN
      ReadString(s);
      Done:=FALSE;
      n:=s[0]='-';
      IF n THEN i:=1 ELSE i:=0 END;
      k:=0;
      REPEAT
        IF (i>5) OR (s[i]<'0') OR (s[i]>'9') THEN RETURN END;
        IF (k>=MAX(CARDINAL) DIV 10) THEN RETURN END;
        k:=10*k+(ORD(s[i])-ORD('0'));
        INC(i);
        IF k>1+CARDINAL(MAX(INTEGER)) THEN RETURN END
      UNTIL s[i]=0C;
      IF n
      THEN
        IF k>CARDINAL(MAX(INTEGER))
        THEN x:=MIN(INTEGER)
        ELSE x:=-INTEGER(k) END;
        Done:=TRUE
      ELSIF k<=CARDINAL(MAX(INTEGER))
      THEN
        x:=INTEGER(k);
        Done:=TRUE
      END
    END ReadInt;

  PROCEDURE ReadReal(VAR x : REAL);
    VAR s : ARRAY [0..20] OF CHAR;
        i, j : CARDINAL;
        n : BOOLEAN;
    BEGIN
      ReadString(s);
      Done:=FALSE;
      x:=0.0;
      n:=s[0]='-';
      IF n THEN i:=1 ELSE i:=0 END;
      REPEAT
        IF (s[i]<'0') OR (s[i]>'9') THEN RETURN END;
        x:=10.0*x+FLOAT(ORD(s[i])-ORD('0'));
        INC(i);
        IF i>=20 THEN RETURN END
      UNTIL (s[i]='.') OR (s[i]=0C);
      IF s[i]<>0C THEN
        INC(i); j:=0;
        REPEAT
          IF (s[i]<'0') OR (s[i]>'9') THEN RETURN END;
          x:=10.0*x+FLOAT(ORD(s[i])-ORD('0'));
          INC(i); INC(j);
          IF i>20 THEN RETURN END
        UNTIL s[i]=0C;
        FOR i:=1 TO j DO x:=x/10.0 END
      END;
      IF n THEN x:=-x END;
      Done:=TRUE
    END ReadReal;

  PROCEDURE Read(VAR c : CHAR);
    BEGIN
      Done:=TRUE;
      c:=ReadChar()
    END Read;

  (* Cursor- und Bildschirm-Steuerung *)

  PROCEDURE GotoXY(x, y : CARDINAL);
    BEGIN
      IF (x>=1) AND (y>=1)
         AND (x<=1+MaxX-MinX)
         AND (y<=1+MaxY-MinY)
      THEN
        CurX:=MinX+(x-1);
        CurY:=MinY+(y-1);
        SetCursor(CurX, CurY)
      END
    END GotoXY;

  PROCEDURE ClrScr;
    VAR X, Y : CARDINAL;
    BEGIN
      GotoXY(1,1);
      FOR Y:=MinY TO MaxY DO
        FOR X:=MinX TO MaxX DO
          ActScreen^[Y,X]:=Color
        END
      END
    END ClrScr;

  PROCEDURE ClrEol;
    VAR X : CARDINAL;
    BEGIN
      FOR X:=CurX TO MaxX DO
        ActScreen^[CurY,X]:=Color
      END
    END ClrEol;

  PROCEDURE InsLine;
    BEGIN
      MoveDown(MinX,CurY,MaxX,MaxY)
    END InsLine;

  PROCEDURE DelLine;
    BEGIN
      MoveUp(MinX,CurY,MaxX,MaxY)
    END DelLine;

  PROCEDURE HighVideo;
    BEGIN
      Color:=0F00H
    END HighVideo;

  PROCEDURE LowVideo;
    BEGIN
      Color:=0700H
    END LowVideo;

  PROCEDURE TextColor(c : CARDINAL);
    BEGIN
      c:=CARDINAL(BITSET(c*100H)*{8..11,15});
      Color:=c + CARDINAL(BITSET(Color)*{12..14})
    END TextColor;

  PROCEDURE TextBackground(c : CARDINAL);
    BEGIN
      c:=CARDINAL(BITSET(c*1000H)*{12..14});
      Color:=c + CARDINAL(BITSET(Color)*{8..11,15})
    END TextBackground;

  PROCEDURE WhereX () : CARDINAL;
    BEGIN
      RETURN 1+CurX-MinX
    END WhereX;

  PROCEDURE WhereY () : CARDINAL;
    BEGIN
      RETURN 1+CurY-MinY
    END WhereY;

  (* Hilfsprozeduren fuer Windows *)

  PROCEDURE MoveScreen(x1,y1,x2,y2 : CARDINAL;
                       VAR S : Screen; VAR TempMem : MemArray);
                       (* sichert Bildschirm *)
    VAR dx, i : CARDINAL;
    BEGIN
      dx:=1+x2-x1;
      FOR i:=0 TO y2-y1 DO
        Move(ADR(S[i+y1,x1]),ADR(TempMem[i*dx]),dx)
      END
    END MoveScreen;

  PROCEDURE RemoveScreen(x1,y1,x2,y2 : CARDINAL;
                         VAR S : Screen; VAR TempMem : MemArray);
                         (* stellt Bildschirminhalt wieder her *)
    VAR dx, i : CARDINAL;
    BEGIN
      dx:=1+x2-x1;
      FOR i:=0 TO y2-y1 DO
        Move(ADR(TempMem[i*dx]),ADR(S[i+y1,x1]),dx)
      END
    END RemoveScreen;

  PROCEDURE len(VAR s : ARRAY OF CHAR) : CARDINAL;
    VAR x : CARDINAL;
    BEGIN
      x:=0;
      WHILE (x<=HIGH(s)) AND (s[x]<>0C) DO INC(x) END;
      RETURN x
    END len;

  PROCEDURE LeaveWindow(w : CARDINAL);
    BEGIN
      WITH WindowArray[w]^ DO
         cx:=WhereX(); cy:=WhereY();
         LastAttr:=Color
      END
    END LeaveWindow;

  PROCEDURE MachWindow(w : CARDINAL);
    BEGIN
      WITH WindowArray[w]^ DO
        Color:=LastAttr;
        MinX:=1+x1;
        MinY:=1+y1;
        MaxX:=x2-1;
        MaxY:=y2-1;
        GotoXY(cx,cy)
      END
    END MachWindow;

  PROCEDURE ZeichneRahmen(x1, y1, x2, y2 : CARDINAL; VAR h : ARRAY OF CHAR);

    VAR i, k, L : CARDINAL;

    PROCEDURE wrtscr(x,y:CARDINAL; c:CHAR);
      BEGIN
        ActScreen^[y,x]:=ORD(c) + 0F00H
      END wrtscr;

    BEGIN
      L:=len(h);
      k:=x1+(x2-x1-L) DIV 2;
      wrtscr(x1,y1,LinksOben);
      FOR i:=x1+1 TO k DO wrtscr(i,y1,WaagrechtO) END;
      FOR i:=k+L+1 TO x2-1 DO
        wrtscr(i,y1,WaagrechtO)
      END;
      wrtscr(x2,y1,RechtsOben);
      FOR i:=y1+1 TO y2-1 DO
          wrtscr(x1,i,SenkrechtL);
          wrtscr(x2,i,SenkrechtR)
      END;
      wrtscr(x1,y2,LinksUnten);
      FOR i:=x1+1 TO x2-1 DO wrtscr(i,y2,WaagrechtU) END;
      wrtscr(x2,y2,RechtsUnten);
      FOR i:=1 TO L DO
        INC(k); wrtscr(k,y1,h[i-1])
      END
    END ZeichneRahmen;

  (*
     Window-Interface:
     Oeffnen und Schliessen der Windows
     sowie Auswahl des aktiven Windows
  *)

  PROCEDURE OpenWindow(x1, y1, x2, y2 : CARDINAL);
    BEGIN
      OpenWindowHeader(x1, y1, x2, y2, '')
    END OpenWindow;

  PROCEDURE OpenWindowHeader(a1, b1, a2, b2 : CARDINAL; h : ARRAY OF CHAR);
    BEGIN
      IF a1<1 THEN a1:=1 END;
      IF b1<1 THEN b1:=1 END;
      IF a1>Spalten-2 THEN a1:=Spalten-2 END;
      IF b1>Zeilen-2  THEN b1:=Zeilen-2 END;
      IF a2>Spalten THEN a2:=Spalten END;
      IF b2>Zeilen  THEN b2:=Zeilen END;
      IF a2<a1+2 THEN a2:=a1+2 END;
      IF b2<b1+2 THEN b2:=b1+2 END;
      IF len(h)>a2-a1-1 THEN h[0]:=0C END;
      IF MaxScreen<>ScreenPtr THEN SelWindow(MaxScreen) END;
      LeaveWindow(ScreenPtr);
      INC(ScreenPtr);
      ALLOCATE(WindowArray[ScreenPtr],SIZE(WindowArray[0]^));
      WITH WindowArray[ScreenPtr]^ DO
        x1:=a1; y1:=b1; x2:=a2; y2:=b2;
        LastAttr:=Color;
        ALLOCATE(ScrPtr,TSIZE(CARDINAL)*(1+a2-a1)*(1+b2-b1));
        MoveScreen(a1,b1,a2,b2,ActScreen^,ScrPtr^);
        ZeichneRahmen(x1, y1, x2, y2, h);
        MachWindow(ScreenPtr);
        ClrScr
      END;
      MaxScreen := ScreenPtr
    END OpenWindowHeader;

  PROCEDURE CloseWindow;
    BEGIN
      IF MaxScreen<>0 THEN
        IF MaxScreen<>ScreenPtr THEN SelWindow(MaxScreen) END;
        WITH WindowArray[ScreenPtr]^ DO
          RemoveScreen(x1,y1,x2,y2,ActScreen^,ScrPtr^);
          DEALLOCATE(ScrPtr,TSIZE(CARDINAL)*(1+x2-x1)*(1+y2-y1))
        END;
        DEALLOCATE(WindowArray[ScreenPtr],SIZE(WindowArray[0]^));
        DEC(ScreenPtr);
        MachWindow(ScreenPtr);
        MaxScreen := ScreenPtr
      END
    END CloseWindow;

  PROCEDURE SelWindow(x : CARDINAL); (* Auswahl bei verdeckten Windows *)
    VAR i          : CARDINAL;

    PROCEDURE Swap(i : CARDINAL);

      BEGIN
        WITH WindowArray[i]^ DO
          MoveScreen(x1,y1,x2,y2,Screencopy,TempMem);
          RemoveScreen(x1,y1,x2,y2,Screencopy,ScrPtr^);
          Move(ADR(TempMem),ScrPtr,(1+x2-x1)*(1+y2-y1))
         END
      END Swap;

    BEGIN
      IF x <= MaxScreen THEN
        LeaveWindow(ScreenPtr);
        Screencopy := ActScreen^;
        IF ScreenPtr = 0
        THEN
          FOR i:=1 TO MaxScreen DO Swap(i) END
        ELSIF ScreenPtr <> MaxScreen THEN
          Swap(ScreenPtr);
          FOR i:=MaxScreen TO 1+ScreenPtr BY -1 DO Swap(i) END;
          FOR i:=ScreenPtr TO MaxScreen DO Swap(i) END
        END;
        IF x = 0
        THEN
          FOR i:=MaxScreen TO 1 BY -1 DO Swap(i) END
        ELSIF x <> MaxScreen THEN
          FOR i:=MaxScreen TO x BY -1 DO Swap(i) END;
          FOR i:=1+x TO MaxScreen DO Swap(i) END;
          Swap(x)
        END;
        ScreenPtr := x;
        ActScreen^ := Screencopy;
        MachWindow(ScreenPtr)
      END
    END SelWindow;

  PROCEDURE ChangeWindow(i : CARDINAL); (* Nur bei nicht verdeckten Windows! *)
    BEGIN
      IF i<=MaxScreen THEN
        LeaveWindow(ScreenPtr);
        ScreenPtr:=i;
        MachWindow(ScreenPtr)
      END
    END ChangeWindow;

BEGIN
  Done:=TRUE;
  LastChar:=0C;
  BufPos:=0; BufEnd:=0;
  MinX:=1; MinY:=1;
  MaxX:=Spalten; MaxY:=Zeilen;

(* Bei Logitech etc. *)
  IF CARDINAL(ScreenMode * BITSET{0..7})=7
  THEN ActScreen:=ADR(MonoScreen)
  ELSE ActScreen:=ADR(ColorScreen) END;

(* Bei Taylor etc.
  ScreenModePtr:=[40H:49H];
  IF ScreenModePtr^*BITSET{0..7}=BITSET{0..2}
  THEN ActScreen:=[0B000H:0] ELSE ActScreen:=[0B800H:0] END; *)

  ScreenPtr:=0; MaxScreen:=0;
  ALLOCATE(WindowArray[0],SIZE(WindowArray[0]^));
  WITH WindowArray[0]^ DO
    x1:=MinX-1; y1:=MinY-1;
    x2:=MaxX+1; y2:=MaxY+1;
    ScrPtr:=NIL
  END;
  LowVideo;
  ClrScr
END ChipIO.