IMPLEMENTATION MODULE COMPLEX;

  FROM MathLib0 IMPORT sin, cos, arctan, sqrt, exp, ln;
  CONST PI=3.1415926536;

  PROCEDURE CMPLX(X, A : REAL; VAR Result : Complex);
    BEGIN
      Result.Re:=cos(A)*X;
      Result.Im:=sin(A)*X;
    END CMPLX;

  PROCEDURE CONJG(X : Complex; VAR Result : Complex);
    BEGIN
      Result.Re:=X.Re;
      Result.Im:=-X.Im;
    END CONJG;

  PROCEDURE CNEG(X : Complex; VAR Result : Complex);
    BEGIN
      Result.Re:=-X.Re;
      Result.Im:=-X.Im;
    END CNEG;

  PROCEDURE CABS(X : Complex) : REAL;
    BEGIN
      RETURN sqrt(X.Re*X.Re+X.Im*X.Im);
    END CABS;

  PROCEDURE ARC(X : Complex) : REAL;
    BEGIN
      IF (X.Re=0.0) AND (X.Im=0.0)
      THEN (* X=0 *)
        RETURN 0.0
      ELSE (* (X.Re<>0.0) OR (X.Im<>0.0) *)
        IF ABS(X.Re)>=ABS(X.Im)
        THEN
          IF X.Re>=0.0
          THEN
            RETURN arctan(X.Im/X.Re);
          ELSE (* X.Re<0.0 *)
            IF X.Im>=0.0
            THEN
              RETURN +PI+arctan(X.Im/X.Re);
            ELSE (* (X.Re<0.0) AND (X.Im<0.0) *)
              RETURN -PI+arctan(X.Im/X.Re);
            END;
          END;
        ELSE (* ABS(X.Re)<ABS(X.Im) *)
          IF X.Im>=0.0
          THEN
            RETURN +PI/2.0-arctan(X.Re/X.Im);
          ELSE
            RETURN -PI/2.0+arctan(X.Re/X.Im);
          END;
        END;
      END;
    END ARC;

  PROCEDURE CADD(A, B : Complex; VAR Result : Complex);
    BEGIN
      Result.Re:=A.Re+B.Re;
      Result.Im:=A.Im+B.Im;
    END CADD;

  PROCEDURE CSUB(A, B : Complex; VAR Result : Complex);
    BEGIN
      Result.Re:=A.Re-B.Re;
      Result.Im:=A.Im-B.Im;
    END CSUB;

  PROCEDURE CMUL(A, B : Complex; VAR Result : Complex);
    BEGIN
      Result.Re:=A.Re*B.Re-A.Im*B.Im;
      Result.Im:=A.Re*B.Im+A.Im*B.Re;
    END CMUL;

  PROCEDURE CDIV(A, B : Complex; VAR Result : Complex);
    VAR Nenner : REAL;
    BEGIN
      Nenner:=B.Re*B.Re+B.Im*B.Im;
      Result.Re:=(A.Re*B.Re+A.Im*B.Im)/Nenner;
      Result.Im:=(A.Im*B.Re-A.Re*B.Im)/Nenner;
    END CDIV;

  PROCEDURE CSQRT(X : Complex; VAR Result : Complex);
    BEGIN
      CMPLX(sqrt(CABS(X)),ARC(X)/2.0,Result);
    END CSQRT;

  PROCEDURE CSIN(X : Complex; VAR Result : Complex);
    VAR e : REAL;
    BEGIN
      e:=exp(X.Im);
      Result.Re:=sin(X.Re)*(e+1.0/e)/2.0;
      Result.Im:=cos(X.Re)*(e-1.0/e)/2.0;
    END CSIN;

  PROCEDURE CCOS(X : Complex; VAR Result : Complex);
    VAR e : REAL;
    BEGIN
      e:=exp(X.Im);
      Result.Re:=cos(X.Re)*(e+1.0/e)/2.0;
      Result.Im:=-sin(X.Re)*(e-1.0/e)/2.0;
    END CCOS;

  PROCEDURE CEXP(X : Complex; VAR Result : Complex);
    BEGIN
      CMPLX(exp(X.Re),X.Im,Result);
    END CEXP;

  PROCEDURE CLOG(X : Complex; VAR Result : Complex);
    BEGIN
      Result.Re:=ln(CABS(X));
      Result.Im:=ARC(X);
    END CLOG;

END COMPLEX.