IMPLEMENTATION MODULE RAND;

  CONST FracPI = 0.1415926535897932385; (* PI-3 *)
  VAR Rnd : REAL;

  PROCEDURE Randomize(StartVal : REAL);
    BEGIN
      Rnd:=StartVal;
      WHILE ABS(Rnd)>=1.0 DO Rnd:=Rnd*FracPI END;
      IF Rnd<0.0 THEN Rnd:=Rnd+1.0 END;
    END Randomize;

  PROCEDURE Random() : REAL;

    (*
       Diese Prozedur berechnet nach dem von HP (Hewlett Packard)
       empfohlenen Algorithmus eine Zufallsfolge.
       Rnd wird dabei auf den gebrochenen Anteil von (Rnd+PI)^8 gesetzt.
       Aufrufbedingung: 0<=Rnd<1
       Endbedingung: 0<=Rnd<1
       Um Ausloeschungsfehler zu vermeiden,
       wird die Berechnung nicht direkt durchgefuehrt,
       sondern getrennt fuer den ganzen und gebrochenen Anteil!
    *)

    VAR G : CARDINAL;  (* Ganzer Anteil von X *)
        R : REAL;      (* Rest von X, 0<=R<1  *)

    PROCEDURE FRAC(X : REAL) : REAL;      (* Gebrochnenen Anteil berechnen  *)
      BEGIN
        RETURN X - FLOAT(TRUNC(X));
      END FRAC;

    PROCEDURE Quad(VAR G : CARDINAL; VAR R : REAL);
      VAR GR : REAL;
      BEGIN
        GR:=FLOAT(2*G)*R;                 (* Gemischtes Glied               *)
        R:=R*R+FRAC(GR);                  (* Gebrochnenen Anteil berechnen  *)
        G:=(G*G)+VAL(CARDINAL,TRUNC(R)+
                     TRUNC(GR));          (* Ganzen Anteil + Uebertrag      *)
        R:=FRAC(R);
      END Quad;

    BEGIN
      R:=Rnd+FracPI;
      G:=3+TRUNC(R);
      R:=FRAC(R);
      Quad(G,R);                          (* (Rnd+PI)^2 berechnen           *)
      Quad(G,R);                          (* (Rnd+PI)^4 berechnen           *)
      Rnd:=FRAC(R*R+FRAC(FLOAT(2*G)*R));  (* NUR FRAC((Rnd+PI)^8) berechnen *)
      RETURN Rnd;                         (* Rnd := "zufaellige" Zahl       *)
    END Random;

BEGIN
  Rnd:=0.0
END RAND.
