IMPLEMENTATION MODULE TEXTFILES;    (* Fr alle Modula-Versionen mit
                                       FileSystem nach Wirth, also
                                       Logitech etc. und auch TopSpeed,
                                       wenn FileSystem implemetiert wird *)

  FROM SYSTEM IMPORT BYTE, ADR;
  FROM Storage IMPORT ALLOCATE, DEALLOCATE;
  IMPORT FileSystem;

  CONST BufLen = 127;

  TYPE TextFile = POINTER TO
                    RECORD
                      F : FileSystem.File;
                      BufPos, BufEnd : CARDINAL;
                      Buffer : ARRAY [0..BufLen] OF CHAR;
                    END;

  PROCEDURE ResetTextFile(VAR f : TextFile; fn : ARRAY OF CHAR);
    BEGIN
      NEW(f);
      FileSystem.Lookup(f^.F,fn,FALSE);
      Done:=(f^.F.res=FileSystem.done);
      IF NOT Done THEN DISPOSE(f); f:=NIL; END;
      FileSystem.SetModify(f^.F);
      f^.BufPos:=0; f^.BufEnd:=0;
    END ResetTextFile;

  PROCEDURE RewriteTextFile(VAR f : TextFile; fn : ARRAY OF CHAR);
    BEGIN
      NEW(f);
      FileSystem.Delete(fn,f^.F);
      FileSystem.Lookup(f^.F,fn,TRUE);
      Done:=(f^.F.res=FileSystem.done);
      IF NOT Done THEN DISPOSE(f); f:=NIL; END;
      FileSystem.SetModify(f^.F);
      f^.BufPos:=0; f^.BufEnd:=0;
    END RewriteTextFile;

  PROCEDURE CloseTextFile(VAR f : TextFile);
    BEGIN
      IF f=NIL THEN Done:=FALSE; RETURN; END;
      FileSystem.Close(f^.F);
      Done:=(f^.F.res=FileSystem.done);
      DISPOSE(f)
    END CloseTextFile;

  PROCEDURE WriteString(VAR f : TextFile; s : ARRAY OF CHAR);
    VAR n, k : CARDINAL;
    BEGIN
      n:=0;
      Done:=TRUE;
      WHILE (n<=HIGH(s)) AND (s[n]<>0C) DO
        INC(n)
      END;
      FileSystem.WriteNBytes(f^.F,ADR(s),n,k);
      Done:=(f^.F.res=FileSystem.done) AND (k=n);
    END WriteString;

  PROCEDURE WriteCard(VAR f : TextFile; x, n : CARDINAL);
    VAR s : ARRAY [0..5] OF CHAR;
        i, j : CARDINAL;
    BEGIN
      i:=0; j:=x;
      REPEAT
        s[i]:=CHR(ORD('0')+j MOD 10);
        j:=j DIV 10;
        INC(i)
      UNTIL (j=0) OR (i>5);
      FOR j:=i+1 TO n DO
        Write(f,' ');
        IF NOT Done THEN RETURN END
      END;
      FOR j:=i-1 TO 0 BY -1 DO
        Write(f,s[j]);
        IF NOT Done THEN RETURN END
      END;
    END WriteCard;

  PROCEDURE WriteInt(VAR f : TextFile; x : INTEGER; n : CARDINAL);
    VAR s : ARRAY [0..6] OF CHAR;
        i, j : CARDINAL;
    BEGIN
      i:=0;
      IF x<>MIN(INTEGER)
      THEN j:=CARDINAL(ABS(x))
      ELSE j:=1+CARDINAL(MAX(INTEGER))
      END;
      REPEAT
        s[i]:=CHR(ORD('0')+j MOD 10);
        j:=j DIV 10;
        INC(i)
      UNTIL (j=0) OR (i>5);
      IF x<0 THEN s[i]:='-'; INC(i) END;
      FOR j:=i+1 TO n DO
        Write(f,' ');
        IF NOT Done THEN RETURN END
      END;
      FOR j:=i-1 TO 0 BY -1 DO
        Write(f,s[j]);
        IF NOT Done THEN RETURN END
      END;
    END WriteInt;

  PROCEDURE WriteReal(VAR f : TextFile; x : REAL; m, n : CARDINAL);
    VAR t : REAL;
        i, j, k : CARDINAL;
    BEGIN
      t:=ABS(x);
      FOR i:=1 TO n DO t:=t*10.0 END;
      t:=t+0.5; j:=1;
      WHILE (t>=10.0) OR (j<=n) DO t:=t/10.0; INC(j) END;
      k:=j;
      IF n<>0 THEN INC(k) END;
      IF x<0.0 THEN INC(k) END;
      FOR i:=k+1 TO m DO
        Write(f,' ');
        IF NOT Done THEN RETURN END
      END;
      IF x<0.0 THEN
        Write(f,'-');
        IF NOT Done THEN RETURN END
      END;
      FOR i:=n+1 TO j DO
        Write(f,CHR(ORD('0')+TRUNC(t)));
        IF NOT Done THEN RETURN END;
        t:=(t-FLOAT(TRUNC(t)))*10.0
      END;
      IF n<>0 THEN
        Write(f,'.');
        IF NOT Done THEN RETURN END
      END;
      FOR i:=1 TO n DO
        Write(f,CHR(ORD('0')+TRUNC(t)));
        IF NOT Done THEN RETURN END;
        t:=(t-FLOAT(TRUNC(t)))*10.0
      END;
    END WriteReal;

  PROCEDURE Write(VAR f : TextFile; c : CHAR);
    BEGIN
      WriteString(f,c)
    END Write;

  PROCEDURE WriteLn(VAR f : TextFile);
    BEGIN
      Write(f,15C);
      IF Done THEN Write(f,12C) END;
    END WriteLn;

  PROCEDURE ReadChar(VAR f : TextFile) : CHAR;
    VAR c : CHAR;
        n : CARDINAL;
    BEGIN
      IF f^.BufPos<f^.BufEnd
      THEN
        c:=f^.Buffer[f^.BufPos];
        INC(f^.BufPos);
        RETURN c
      ELSE
        f^.BufEnd:=0;
        REPEAT
          FileSystem.ReadNBytes(f^.F,ADR(c),1,n);
          Done:=(f^.F.res=FileSystem.done) AND (n=1);
          IF NOT Done THEN c:=15C END;
          IF (c>=' ') AND (f^.BufEnd<BufLen)
          THEN
            f^.Buffer[f^.BufEnd]:=c;
            INC(f^.BufEnd);
          END
        UNTIL c<' ';
        f^.Buffer[f^.BufEnd]:=15C;
        INC(f^.BufEnd);
        f^.BufPos:=1;
        RETURN f^.Buffer[0]
      END
    END ReadChar;

  PROCEDURE ReadLn(VAR f : TextFile);
    BEGIN
      f^.BufPos:=f^.BufEnd
    END ReadLn;

  PROCEDURE ReadString(VAR f : TextFile; VAR s : ARRAY OF CHAR);
    VAR c : CHAR;
        i : CARDINAL;
    BEGIN
      Done:=TRUE;
      REPEAT
        c:=ReadChar(f)
      UNTIL (c>' ') OR NOT Done;
      i:=0;
      IF Done THEN
        REPEAT
          s[i]:=c;
          INC(i);
          IF i>HIGH(s) THEN RETURN END;
          c:=ReadChar(f)
        UNTIL (c<=' ') OR NOT Done;
      END;
      s[i]:=0C
    END ReadString;

  PROCEDURE ReadCard(VAR f : TextFile; VAR x : CARDINAL);
    VAR s : ARRAY [0..5] OF CHAR;
        i : CARDINAL;
    BEGIN
      ReadString(f,s);
      IF NOT Done THEN RETURN END;
      Done:=FALSE;
      x:=0;
      i:=0;
      REPEAT
        IF (i>4) OR (s[i]<'0') OR (s[i]>'9') THEN RETURN END;
        IF (x>MAX(CARDINAL) DIV 10) OR
           (10*x>MAX(CARDINAL)-(ORD(s[i])-ORD('0'))) THEN RETURN END;
        x:=10*x+(ORD(s[i])-ORD('0'));
        INC(i)
      UNTIL s[i]=0C;
      Done:=TRUE
    END ReadCard;

  PROCEDURE ReadInt(VAR f : TextFile; VAR x : INTEGER);
    VAR s : ARRAY [0..6] OF CHAR;
        i, k : CARDINAL;
        n : BOOLEAN;
    BEGIN
      ReadString(f,s);
      IF NOT Done THEN RETURN END;
      Done:=FALSE;
      n:=s[0]='-';
      IF n THEN i:=1 ELSE i:=0 END;
      k:=0;
      REPEAT
        IF (i>5) OR (s[i]<'0') OR (s[i]>'9') THEN RETURN END;
        IF (k>=MAX(CARDINAL) DIV 10) THEN RETURN END;
        k:=10*k+(ORD(s[i])-ORD('0'));
        INC(i);
        IF k>1+CARDINAL(MAX(INTEGER)) THEN RETURN END
      UNTIL s[i]=0C;
      IF n
      THEN
        IF k>CARDINAL(MAX(INTEGER))
        THEN x:=MIN(INTEGER)
        ELSE x:=-INTEGER(k) END;
        Done:=TRUE
      ELSIF k<=CARDINAL(MAX(INTEGER))
      THEN
        x:=INTEGER(k);
        Done:=TRUE
      END
    END ReadInt;

  PROCEDURE ReadReal(VAR f : TextFile; VAR x : REAL);
    VAR s : ARRAY [0..20] OF CHAR;
        i, j : CARDINAL;
        n : BOOLEAN;
    BEGIN
      ReadString(f,s);
      IF NOT Done THEN RETURN END;
      Done:=FALSE;
      x:=0.0;
      n:=s[0]='-';
      IF n THEN i:=1 ELSE i:=0 END;
      REPEAT
        IF (s[i]<'0') OR (s[i]>'9') THEN RETURN END;
        x:=10.0*x+FLOAT(ORD(s[i])-ORD('0'));
        INC(i);
        IF i>=20 THEN RETURN END
      UNTIL (s[i]='.') OR (s[i]=0C);
      IF s[i]<>0C THEN
        INC(i); j:=0;
        REPEAT
          IF (s[i]<'0') OR (s[i]>'9') THEN RETURN END;
          x:=10.0*x+FLOAT(ORD(s[i])-ORD('0'));
          INC(i); INC(j);
          IF i>20 THEN RETURN END
        UNTIL s[i]=0C;
        FOR i:=1 TO j DO x:=x/10.0 END
      END;
      IF n THEN x:=-x END;
      Done:=TRUE
    END ReadReal;

  PROCEDURE Read(VAR f : TextFile; VAR c : CHAR);
    BEGIN
      Done:=TRUE;
      c:=ReadChar(f)
    END Read;

BEGIN
  Done:=TRUE;
END TEXTFILES.